#
# Copyright 2019 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inaccel.coral as inaccel
import numpy as np
import time

class CFBlackScholes:
	NUM_ELEMENTS_PER_BUFFER_CHUNK = 16;

	def __init__(self, maxNumAssets):
		self.numAssets = maxNumAssets
		self.allocateBuffers(maxNumAssets)

	def allocateBuffers(self, numRequestedElements):
		m_numPaddedBufferElements = self.calculatePaddedNumElements(numRequestedElements)

		with inaccel.allocator:
			self.stockPrice = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.strikePrice = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.volatility = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.riskFreeRate = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.timeToMaturity = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)

			self.optionPrice = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)

			self.delta = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.gamma = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.vega = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.theta = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)
			self.rho = np.ndarray(m_numPaddedBufferElements, dtype = np.float32)

	def calculatePaddedNumElements(self, numRequestedElements):
		numChunks = int((numRequestedElements + (self.NUM_ELEMENTS_PER_BUFFER_CHUNK - 1)) / self.NUM_ELEMENTS_PER_BUFFER_CHUNK)

		numPaddedElements = numChunks * self.NUM_ELEMENTS_PER_BUFFER_CHUNK

		return numPaddedElements

	def run(self, optionType, numAssets):
		self.runAsync(optionType, numAssets)
		self.wait()

	def runAsync(self, optionType, numAssets):
		self.m_runStartTime = int(round(time.time() * 1000000))

		numPaddedAssets = np.int32(self.calculatePaddedNumElements(numAssets))

		req = inaccel.request("com.xilinx.vitis.quantitativeFinance.blackScholes.calculator")

		req.arg(self.stockPrice).arg(self.volatility).arg(self.riskFreeRate).arg(self.timeToMaturity).arg(self.strikePrice).arg(optionType.value).arg(numPaddedAssets).arg(self.optionPrice).arg(self.delta).arg(self.gamma).arg(self.vega).arg(self.theta).arg(self.rho)

		self.response = inaccel.submit(req)

	def wait(self):
		self.response.result()

		self.m_runEndTime = int(round(time.time() * 1000000))

	def lastruntime(self):
		duration = self.m_runEndTime - self.m_runStartTime

		return duration



import inaccel.coral as inaccel
import numpy as np
import time

class StereoBM:
	def __init__(self, cameraMA_l=None, cameraMA_r=None, distC_l=None, distC_r=None, irA_l=None, irA_r=None, bm_state=None ):
		# allocate mem for camera parameters for rectification and bm_state class
		with inaccel.allocator:
			if cameraMA_l is None:
				self.cameraMA_l_fl = np.array([933.173, 0.0, 663.451, 0.0, 933.173, 377.015, 0.0, 0.0, 1.0], dtype=np.float32)
			else:
				self.cameraMA_l_fl = np.array(cameraMA_l, dtype=np.float32)

			if cameraMA_r is None:
				self.cameraMA_r_fl = np.array([933.467, 0.0, 678.297, 0.0, 933.467, 359.623, 0.0, 0.0, 1.0], dtype=np.float32)
			else:
				self.cameraMA_r_fl = np.array(cameraMA_r, dtype=np.float32)

			if distC_l is None:
				self.distC_l_fl = np.array([-0.169398, 0.0227329, 0.0, 0.0, 0.0], dtype=np.float32)
			else:
				self.distC_l_fl = np.array(distC_l, dtype=np.float32)

			if distC_r is None:
				self.distC_r_fl = np.array([-0.170581, 0.0249444, 0.0, 0.0, 0.0], dtype=np.float32)
			else:
				self.distC_r_fl = np.array(distC_r, dtype=np.float32)

			if irA_l is None:
				self.irA_l_fl = np.array([0.0011976323, -0.0000000019, -0.8153011732, 0.0000000007, 0.0011976994, \
	                      								-0.4422348617,  0.0000126839,  0.0000001064, 0.9913820905], dtype=np.float32)
			else:
				self.irA_l_fl = np.array(irA_l, dtype=np.float32)

			if irA_r is None:
				self.irA_r_fl = np.array([0.0011976994,  0.0000000000, -0.8047567905, -0.0000000000, 0.0011976994, \
	                      -0.4420566166, -0.0000000000, -0.0000001064,  1.0000392898], dtype=np.float32)
			else:
				self.irA_r_fl = np.array(irA_r, dtype=np.float32)

			if bm_state is None:
				self.bm_state_arr = np.array([0, 15, 31, 15, 0, 48, 20, 15, 16, 3, 0], dtype=np.int32)
			else:
				self.bm_state_arr = np.array(bm_state, dtype=np.int32)


	def runAsync(self, left_img, right_img):
		self.m_runStartTime = int(round(time.time() * 1000000))

		if left_img is None:
			raise RuntimeError('Invalid left image')
		if right_img is None:
			raise RuntimeError('Invalid right image')
		if left_img.shape[0] != right_img.shape[0] or left_img.shape[1] != right_img.shape[1]:
			raise RuntimeError('Image sizes differ')

		# allocate and initialize buffers
		rows = np.int32(left_img.shape[0]);
		cols = np.int32(left_img.shape[1]);

		with inaccel.allocator:
			self.left_mat = np.array(left_img)
			self.right_mat = np.array(right_img)
			self.disp_mat = np.ndarray((rows, cols), dtype=np.uint16)

		# Create request for stereo accelerator
		req = inaccel.request('com.xilinx.vitis.vision.stereoBM')
		req.arg(self.left_mat).arg(self.right_mat).arg(self.disp_mat)
		req.arg(self.cameraMA_l_fl).arg(self.cameraMA_r_fl)
		req.arg(self.distC_l_fl).arg(self.distC_r_fl)
		req.arg(self.irA_l_fl).arg(self.irA_r_fl)
		req.arg(self.bm_state_arr)
		req.arg(rows).arg(cols)

		self.response = inaccel.submit(req)

	def wait(self):
		# Send request and wait for completion
		self.response.result()

		# Write output image
		disp_mat_scaled = (self.disp_mat.view(np.ndarray)*(256.0 / 48.0) / (16.0)).astype(np.uint8)

		self.m_runEndTime = int(round(time.time() * 1000000))
		return disp_mat_scaled;

	def run(self, left_img, right_img):
		self.runAsync(left_img, right_img)
		return self.wait()

	def lastruntime(self):
		duration = self.m_runEndTime - self.m_runStartTime
		return duration



from .xf_fintech_binomialtree import BinomialTree
from .xf_fintech_cf_black_scholes import CFBlackScholes
from .xf_fintech_mc_american import MCAmerican
from .xf_fintech_types import OptionType

__all__ = ["BinomialTree", "CFBlackScholes","MCAmerican","OptionType"]



import cv2 as cv
import numpy as np
import math
import time

from ._vision_stereobm import StereoBM


class StereoBM_video:

	def __init__(self, parallel_requests=None):
		self.stereo = []
		self.parallel_requests = 16
		if parallel_requests is not None:
			self.parallel_requests = parallel_requests

		for i in range(0, self.parallel_requests):
			self.stereo.append( StereoBM() )

	def run(self, input_video, output_video, split_dimention=0):
		self.m_runStartTime = int(round(time.time() * 1000000))

		# try to open the video
		vid = cv.VideoCapture(input_video)
		if (vid.isOpened()== False):
			  raise RuntimeError("Error opening video stream or file")

		frame_num = int(vid.get(cv.CAP_PROP_FRAME_COUNT))
		if(split_dimention == 0):
			height = int(vid.get(cv.CAP_PROP_FRAME_HEIGHT))
			width  = int(vid.get(cv.CAP_PROP_FRAME_WIDTH)/2)
		else:
			height = int(vid.get(cv.CAP_PROP_FRAME_HEIGHT)/2)
			width  = int(vid.get(cv.CAP_PROP_FRAME_WIDTH))

		fourcc = int(vid.get(cv.CAP_PROP_FOURCC))
		fps = int(vid.get(cv.CAP_PROP_FPS))
		out = cv.VideoWriter(output_video, fourcc, fps, (height,width))

		frames = []
		for i in range(math.ceil(frame_num/self.parallel_requests)):
			frames.clear()

			for j in range(self.parallel_requests):
				ret, frame = vid.read()
				if ret == False:
					break
				else:
					frames.append(frame)

			for j in range(len(frames)):
				if(split_dimention == 0):
					frame_left = cv.cvtColor(cv.resize(frames[j][:,:width,:], (1280, 720)), cv.COLOR_BGR2GRAY)
					frame_right = cv.cvtColor(cv.resize(frames[j][:,width:,:], (1280, 720)), cv.COLOR_BGR2GRAY)
				else:
					frame_left = cv.cvtColor(cv.resize(frames[j][:height,:,:], (1280, 720)), cv.COLOR_BGR2GRAY)
					frame_right = cv.cvtColor(cv.resize(frames[j][height:,:,:], (1280, 720)), cv.COLOR_BGR2GRAY)

				self.stereo[j].runAsync(frame_left, frame_right)

			for j in range(len(frames)):
				disp_img = self.stereo[j].wait()
				disp_img = cv.resize(cv.cvtColor(disp_img, cv.COLOR_GRAY2BGR), (height,width))
				out.write(disp_img)

		vid.release()
		out.release()
		self.m_runEndTime = int(round(time.time() * 1000000))

	def lastruntime(self):
		duration = self.m_runEndTime - self.m_runStartTime
		return duration



#
# Copyright 2019 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inaccel.coral as inaccel
import numpy as np
import time

class MCAmerican:
	UN_K1 = 2
	COEF = 4
	TIMESTEPS = 100
	ITERATION = 4
	SZ = 8 * np.double().nbytes

	PRICE_ELEMENT_SIZE = np.double().nbytes * UN_K1
	MATRIX_ELEMENT_SIZE = np.double().nbytes
	COEFF_ELEMENT_SIZE = np.double().nbytes * COEF

	PRICE_NUM_ELEMENTS = 1024 * TIMESTEPS * ITERATION
	MATRIX_NUM_ELEMENTS = 9 * TIMESTEPS
	COEFF_NUM_ELEMENTS = TIMESTEPS - 1

	def __init__(self):
		self.flag = True
		with inaccel.allocator:
			self.m_hostOutputPricesBuffer = np.ndarray(self.PRICE_ELEMENT_SIZE * self.PRICE_NUM_ELEMENTS, dtype = np.uint8)
			self.m_hostOutputMatrixBuffer = np.ndarray(self.MATRIX_ELEMENT_SIZE * self.MATRIX_NUM_ELEMENTS, dtype = np.uint8)
			self.m_hostCoeffBuffer = np.ndarray(self.COEFF_ELEMENT_SIZE * self.COEFF_NUM_ELEMENTS, dtype = np.uint8)
			self.m_hostOutputBuffer1 = np.ndarray(1, dtype = np.double)
			self.m_hostOutputBuffer2 = np.ndarray(1, dtype = np.double)

	def run(self, optionType, _stockPrice, _strikePrice, _riskFreeRate, _dividendYield, _volatility, _timeToMaturity, samplesOrTolerance):
		self.m_runStartTime = int(round(time.time() * 1000000))

		stockPrice = np.double(_stockPrice)
		strikePrice = np.double(_strikePrice)
		riskFreeRate = np.double(_riskFreeRate)
		dividendYield = np.double(_dividendYield)
		volatility = np.double(_volatility)
		timeToMaturity = np.double(_timeToMaturity)

		if isinstance(samplesOrTolerance, float):
			requiredTolerance = np.double(samplesOrTolerance)
			requiredSamples = np.int32(0)
		else:
			requiredTolerance = np.double(0.0)
			requiredSamples = np.int32(samplesOrTolerance)

		calibrateSamples = np.int32(4096)

		timeSteps = np.int32(self.TIMESTEPS)

		# --------------------
		# Run PRESAMPLE kernel
		# --------------------
		preSample = inaccel.request("com.xilinx.vitis.quantitativeFinance.monteCarlo.PreSample")

		preSample.arg(stockPrice).arg(volatility).arg(riskFreeRate).arg(dividendYield).arg(timeToMaturity).arg(strikePrice).arg(optionType.value).arg(self.m_hostOutputPricesBuffer).arg(self.m_hostOutputMatrixBuffer).arg(calibrateSamples).arg(timeSteps)

		# ----------------------
		# Run CALIBRATION kernel
		# ----------------------
		calibration = inaccel.request("com.xilinx.vitis.quantitativeFinance.monteCarlo.Calibration");
		calibration.arg(timeToMaturity).arg(riskFreeRate).arg(strikePrice).arg(optionType.value).arg(self.m_hostOutputPricesBuffer).arg(self.m_hostOutputMatrixBuffer).arg(self.m_hostCoeffBuffer).arg(calibrateSamples).arg(timeSteps);

		# -------------------
		# Run PRICING kernels
		# -------------------
		pricing1 = inaccel.request("com.xilinx.vitis.quantitativeFinance.monteCarlo.Pricing1");
		pricing1.arg(stockPrice).arg(volatility).arg(dividendYield).arg(riskFreeRate).arg(timeToMaturity).arg(strikePrice).arg(optionType.value).arg(self.m_hostCoeffBuffer).arg(self.m_hostOutputBuffer1).arg(requiredTolerance).arg(requiredSamples).arg(timeSteps);

		inaccel.submit(preSample).result()
		inaccel.submit(calibration).result()
		prcng1 = inaccel.submit(pricing1)

		if (self.flag):
			try:
				pricing2 = inaccel.request("com.xilinx.vitis.quantitativeFinance.monteCarlo.Pricing2")
				pricing2.arg(stockPrice).arg(volatility).arg(dividendYield).arg(riskFreeRate).arg(timeToMaturity).arg(strikePrice).arg(optionType.value).arg(self.m_hostCoeffBuffer).arg(self.m_hostOutputBuffer2).arg(requiredTolerance).arg(requiredSamples).arg(timeSteps)

				inaccel.submit(pricing2).result()
			except Exception:
				self.flag = False

		prcng1.result()

		self.m_runEndTime = int(round(time.time() * 1000000))

		# ----------------------------------------------------------------------------------------
		# Average the outputs from the two pricing kernels, and give the result
		# back to the caller
		# ----------------------------------------------------------------------------------------
		if (self.flag):
			return (self.m_hostOutputBuffer1[0] + self.m_hostOutputBuffer2[0]) / 2.0
		else:
			return self.m_hostOutputBuffer1[0]

	def lastruntime(self):
		duration = self.m_runEndTime - self.m_runStartTime

		return duration



"""Setup inaccel-vitis package."""
from setuptools import find_namespace_packages, setup

setup(
    name = 'inaccel-vitis',
    packages = find_namespace_packages(include = ['inaccel.*']),
    namespace_packages = ['inaccel'],
    version = '0.2',
    license = 'Apache-2.0',
    description = 'InAccel Vitis Libraries',
    author = 'InAccel',
    author_email = 'info@inaccel.com',
    url = 'https://docs.inaccel.com',
    keywords = ['InAccel Coral', 'FPGA', 'inaccel', 'Vitis'],
    install_requires = [
        'coral-api==2.*', 'opencv-python',
    ],
    include_package_data = True,
    zip_safe = True,
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires = '>=3.8',
)



#
# Copyright 2019 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inaccel.coral as inaccel
import numpy as np
import time

BinomialTreeInputDataTypeDouble = np.dtype([('S', np.double), ('K', np.double), ('T', np.double), ('rf', np.double), ('V', np.double), ('q', np.double), ('N', np.int32), ('packed', np.int32, 3)])

class BinomialTree:
	MAX_OPTION_CALCULATIONS = 1024;

	BinomialTreeEuropeanPut = np.int32(1);
	BinomialTreeEuropeanCall = np.int32(2);
	BinomialTreeAmericanPut = np.int32(3);
	BinomialTreeAmericanCall = np.int32(4);

	def __init__(self):
		with inaccel.allocator:
			self.inputBuffer = np.ndarray(self.MAX_OPTION_CALCULATIONS, dtype = BinomialTreeInputDataTypeDouble)
			self.outputBuffer = np.ndarray(self.MAX_OPTION_CALCULATIONS, dtype = np.double)

	def run(self, optionType):
		self.m_runStartTime = int(round(time.time() * 1000000))

		numOptions = np.int32(self.inputBuffer.size)
		startIndex = np.int32(0)

		if ((numOptions % 8) != 0):
			raise RuntimeError("[XLNX] BinomialTree::run - number of options to calculate should be a multiple of 8")

		req = inaccel.request("com.xilinx.vitis.quantitativeFinance.binomialTree.engine")

		req.arg(self.inputBuffer).arg(self.outputBuffer).arg(optionType).arg(numOptions).arg(startIndex)

		inaccel.submit(req).result()

		self.m_runEndTime = int(round(time.time() * 1000000))

	def lastruntime(self):
		duration = self.m_runEndTime - self.m_runStartTime

		return duration
